$!								2-JUN-1994
$!
$! This command procedure runs the http_server.exe image and must reside
$! in the same directory.  It takes the following parameters:
$!
$!	Param.	  Default		Description.
$!	P1	RUN		Mode option:
$!				   RUN		Run server image.
$!				   DETACH	Run server in detached process.
$!				   INSTALL	Install server image and
$!						DECthreads shareables.
$!
$!      P2  -	HTTP_SERVER.LOG	Log file name.  File will default to same
$!				directory as this procedure.
$!
$!      P3  -   HTTP_SERVER.CONF Configuration file name.  This file is also
$!				known as the rules file.  File will default to
$!				same directory as this procedure.
$!
$!	P4  -    80		TCP port number that server will listen on.
$!
$!	P5  -    		TCP port number that server will listen on
$!				with cache disabled.
$!
$! To start the server under another username (i.e. the WWW server account),
$! from system startup, submit this procedure as a batch job with the
$! appropriate /USERNAME and /PARAMETER qualifiers.
$!
$! Note that if http_server.exe is to be installed with privilege, then you 
$! must also install sys$share:cma$lib_shr.exe, sys$share:cma$open_rtl.exe, 
$! and sys$share:cma$rtl.exe.  Invoking this procedure with a P1 parameter of
$! "INSTALL" will install these images.
$!
$ proc = f$parse(f$environment("PROCEDURE"),,,,"NO_CONCEAL")
$ if P1 .eqs. "" THEN P1 = "RUN"
$ if P1 .nes. "RUN" .and. P1 .nes. "DETACH" .and. P1 .nes. "INSTALL"
$ then
$    write sys$output "Invalid P1 parameter, must be RUN, DETACH, or INSTALL"
$    exit 20
$ endif
$!
$ if P1 .eqs. "DETACH"
$ then
$     outfile = f$parse(P2,"HTTP_SERVER.OUT;",proc)
$!
$!    Start detached process with same quotas as current process.
$!
$     run sys$system:loginout/detach/input='proc'/output='outfile' -
	/buffer_limit='f$getjpi("0","bytlm")' -
	/file_limit='f$getjpi("0","fillm")' -
	/io_buffered='f$getjpi("0","biolm")' -
	/ast_limit='f$getjpi("0","astlm")' -
	/page_file='f$getjpi("0","pgflquota")' -
	/process_name="HTTP Server"-
	/priority=4
$     exit $status
$ else
$   if P1 .eqs. "INSTALL"
$   then
$!
$!      Install http image with privilege needed to listen on privileged
$!	TCP port (UCX), which requires that DECthreads images be installed
$!	also.
$!
$	image = f$parse(".EXE;",proc) - ";"
$	privs = "OPER,PHY_IO"
$	if f$trnlnm("MULTINET") .nes. "" then privs = "SYSPRV,PHY_IO"
$	install replace 'image'/privilege=('privs')
$	if .not. f$file("sys$share:cma$lib_shr.exe","known") then -
	    install replace sys$share:cma$lib_shr.exe
$!
$	if .not. f$file("sys$share:cma$open_rtl.exe","known") then -
	    install replace sys$share:cma$open_rtl.exe
$!
$	if .not. f$file("sys$share:cma$rtl.exe","known") then -
	    install replace sys$share:cma$rtl.exe
$!
$	if f$getsyi("CPU") .ge. 128		! test for alpha
$	then
$	    if .not. f$file("sys$share:cma$open_lib_shr.exe","known") then -
	        install replace sys$share:cma$open_lib_shr.exe
$	endif
$	exit
$   endif
$ endif
$!
$! Define DCL symbols that HTTP_SERVER will read as environment variables.
$!
$ HTTP_CACHE_SIZE = 0			! Entries in small document cache.
$ HTTP_CLIENT_LIMIT = 100		! Max. number of concurent clients.
$ HTTP_LOG_LEVEL = 3			! Level of detail in log file.
$!
$! Make default for default host name, this will work with both UCX and
$! multinet.
$!
$ if f$type(http_default_host) + f$trnlnm("HTTP_DEFAULT_HOST") .eqs. "" 
$ then
$	hname = f$trnlnm("UCX$INET_HOST")
$	if f$locate(".",hname) .ge. f$length(hname) then -
		hname = hname + "." + f$trnlnm("UCX$INET_DOMAIN")
$ 	write sys$output "Using default host name of ",hname
$	HTTP_DEFAULT_HOST = hname
$ endif
$!
$! Define www_root if not defined.
$!
$ if f$trnlnm("www_root") .eqs. ""
$ then
$    root_dir = f$parse("1.;",proc,,,"NO_CONCEAL,SYNTAX_ONLY") - "][" - "]1.;"
$    root_dir = root_dir + ".]"
$    write sys$output "Defining www_root as ", root_dir
$    define www_root 'root_dir'/trans=(term,conceal)
$ endif
$!
$! Setup command line arguments, Make this procedure name the related file
$! to default it to the same directory.
$!
$ log_file = f$parse(P2,".LOG",proc)
$ rules_file = f$parse(P3,".CONF",proc)
$ port = "80"
$ if P4 .nes. "" then port = P4
$!
$! Run server as foreign command.
$!
$ httpd = "$" + f$parse(".EXE;",proc) - ";"
$ httpd 'log_file' 'rules_file' 'port' 'P5' 'P6'
$ exit $status
